/*
plot3d - three dimensional plotting routine 
*/	

#ident "$Id: plot3d.c,v 1.7 2007/05/22 15:33:04 ccrawford Exp $"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include "crcplot.h"

				/* initial upper elevation value; must
				   be out of range. drawline keys on 
				   this value to see that no lines have
				   been drawn through a point */
#define ZINIT	-123456789		

#define HEIGHT 0.14		/* ... of top and bottom labels */

#define	Z(A,B)	(z[(B)*nx+(A)]) /* easy access to elevation array */

static float pxmin,pxmax;	/* physical x extent from bounding box */
static float pymin,pymax;	/* physical y extent from bounding box */
static float pyminb,pymaxb;	/* physical y extent from data */
static int numstp;		/* number of points in elevation arrays */
static int *above,*below;	/* elevation arrays */

				/* scale factor from discrete space back to
				   physical space. used by iplot */
static float iplot_scale;	

				/* the following variables are used by 
				   xcal, ycal, and dcal to transform 
				   object space into physical space.
				   the functions are used once to 
				   determine the projection of the bounding
				   box. they are used a second time to
				   perform the transformations. the
				   parameters have to be adjusted to 
				   perform the two different functions */

static float absxs,absys,s1s2,sphi1,cphi1,cphi2,c1s2,c2s1,c1c2,sphi2;

static int *absx,*absy;		/* locations for 'current' image line */
static int *xold,*yold;		/* locations for previous image line */

static int debug = 0;		/* debug level */

/*
	xcal - calculate the absolute x (image space) coordinate of a point 
*/
static float xcal(x1,y1)
float x1,y1;
{
	return((cphi1 * x1 + sphi1 * y1 - absxs) );
}

/*
	ycal - calculate the absolute y (image space) coordinate of a point 	
*/
static float ycal(x1,y1,z1)
float x1,y1,z1;
{
	return((-s1s2 * x1 + c1s2 * y1 + cphi2 * z1 - absys));
}

/*
	dcal - calculate the depth (image space) coordinate of a point 	
*/
static float 
dcal(x1,y1,z1)
float x1,y1,z1;
{
	return((-c2s1 * x1 + c1c2 * y1 + sphi2 * z1));
}

/*
	find max of two floats
*/
static float 
max(a,b)
	float a,b;
{
	return((a > b)? a : b);
}

/*
	find min of two floats
*/
static float 
min(a,b)
	float a,b;
{
	return((a < b)? a : b);
}


/*
	print error message and exit
*/
static void 
err(s)
char *s;
{
	fputs("plot3d: ",stderr);
	fputs(s,stderr);
	fputs("\n",stderr);
	exit(1);
}

/*
	iplot - integer interface to plot
*/
static void 
iplot(xx,yy,m)
int xx,yy,m;
{
	plot(iplot_scale * xx,iplot_scale * yy,m);
}

/*
	drawline - draw a line and set the hiding boundaries accorindingly
*/
static void 
drawline(x1,y1,x2,y2)
int x1,y1,x2,y2;
{
	int m;		/* equivalent to x location */
	int abv,blw=-1;	/* 1=above (below) upper (lower) horizon */
	int dx,dy;	/* x2 - x1, y2 - y1 */
	int ix,iy;	/* used to set up bresenham */
	int majx,majy;	/* bresenham major step */
	int minx,miny;	/* bresenham minor step */
	int nb,na,nc,nctr;	/* counters used by bresemham */
	int steep;	/* 1=steep line */
	int xpos,ypos;	/* start of line for potential plotting */
	int by=-1,ay=-1;	/* original horizons, used in steep algorithm */
	int fp;		/* 0=first point (section) used by steep */

			/* figure out how much to move */
	dx = x2 - x1;
	dy = y2 - y1;

			/* don't plot zero length lines */
	if(dx == 0 && dy == 0)return;

			/* keep track of tail of segment for potential 
			later plotting */
	xpos = x1;
	ypos = y1;

			/* bresenham, resolve to first quadrant */
	ix=1;
	if(dx<0){
		ix = -1;
		dx = -dx;
	}
	iy=1;
	if(dy<0){
		iy = -1;
		dy = -dy;
	}

			/* bresenham, resolve to proper octant */
	majy = iy;
	majx = ix;
	if(dx<=dy){
		steep = 1;
		minx = 0;
		miny = iy;
		na = dy;
		nb = dx + dx;
	}
	else{
		steep = 0;
		miny = 0;
		minx = ix;
		na = dx;
		nb = dy + dy;
	}
	nc = na + na;
	nctr = na;

			/* let steep algorithm know its first point */
	fp = 0;

			/* m is a synonym for x for most of code */
	m = x1;

			/* set initial state of line */
	if(above[m] == ZINIT){
		abv = 1;
	}else{
		abv = y1 > above[m];
		blw = y1 < below[m];
	}

			/* for steep, save original horizons */
	if(steep){
		ay = above[m];
		if(above[m] != ZINIT) by = below[m];
	}

			/* loop through discrete samples */
	while(1){
			/* this check is unnecessary for debugged code */
		if((m < 0) || (m > numstp)) err("exceeded hiding boundary");

			
			/* check for uninitialed horizon */
		if(above[m] == ZINIT){
			below[m] = above[m] = y1;
			abv = 1;
		}else{ 
				/* line starts out above upper horizon */
			if(abv){
				if(y1 <= above[m]){
					iplot(xpos,ypos,3);
					iplot(x1,y1,2);
					abv = 0;
						/* might not be necessary */
					if(y1 < below[m]){
						blw = 1;
						below[m] = y1;
					}
				}
				/* line starts out below lower horizon */
			}else if(blw){
				if(y1 >= below[m]){
					iplot(xpos,ypos,3);
					iplot(x1,y1,2);
					blw = 0;
						/* might not be necessary */
					if(y1 >= above[m]){
						abv = 1;
						above[m] = y1;
					}
				}
				/* line between horizons */
			}else{
					/* above upper horizon? */
				if(y1 >= above[m]){
					xpos = x1;
					ypos = y1;
					abv = 1;
				}
				
					/* below lower horizon */
				if(y1 <= below[m]){
					xpos = x1;
					ypos = y1;
					blw = 1;
				}
			}

				/* update horizons */
			if(!steep || fp){
				if(y1 < below[m]){
					below[m] = y1; 
					blw = 1;
				}
				
				if(y1 > above[m]){
					above[m] = y1; 
					abv = 1;
				}
			}
		}

			/* step to next x location; steep code needs to 
			loop at this point and do checks while looping */
it:
		na = na + nb;
		if(na>=nc){
			na -= nc;
			x1 += majx;
			y1 += majy;
		}
		else{
			x1 += minx;
			y1 += miny;
		}
		nctr -= 1;

			/* break out of major while loop */
		if(nctr <= 0) break;

		if(steep){
				/* at same x? */
			if(x1 == m){
				if(ay == ZINIT){
					goto it;
				}else if(abv){
					if(!fp || y1 > ay){
						goto it;
					}
				}else if(blw){
					if(!fp || y1 < by){
						goto it;
					}
				}else{
					if(y1 > above[m]){
						abv = 1;
					}else if(y1 < below[m]){
						blw = 1;
					}else if(y1!=below[m] && y1!=above[m]){
						goto it;
					}
					xpos = x1;
					ypos = y1;
				}
				/* at new x */
			}else{
				m = x1;
				if(!abv && !blw){
					if(y1 <= below[m]){
						blw = 1;
						if(fp){
							xpos = x1;
							ypos = y1;
						}
					}
					
					if(y1 >= above[m]){
						abv = 1;
						if(fp){
							xpos = x1;
							ypos = y1;
						}
					}
				}
			}
		}

			/* prepare for next x location */
		m = x1;
		by = below[m];
		ay = above[m];
		fp = 1;
	}

		/* plot [remaining] segments that didn't intersect */
	if(abv || blw ){
		iplot(xpos,ypos,3);
		iplot(x1,y1,2);
	}

		/* update horizons */
	if(above[x1] == ZINIT){
		above[x1] = below[x1] = y1;
	}else{
		if(y1 > above[x1])above[x1] = y1;
		if(y1 < below[x1])below[x1] = y1;
	}

		/* check not necessary for debugged code */
	if(x1 != x2 || y2 != y2)err("x/y mismatch");
}

/*
	perpen - draws the small lines between two adjacent stripes	
*/
static void 
perpen(index)
int index;
{
	drawline(xold[index],yold[index],absx[index],absy[index]);
}


/*
	drawaxes - draw rotated axis (interface to axis)
*/
static void
drawaxes(x1,y1,x2,y2,vmin,vmax,dir,type,label)

			/* end points of axis, relative to bounding box */
float x1,y1,x2,y2;

			/* vector minimum and maximum */
float vmin,vmax;

			/* 1=switch vmin and vmax; depends on which 
			corner axis originates from */
int dir;
			/* axis type; 2= z axis; 0=lower left; 1= lower right */
int type;

			/* axis label */
char *label;
{
	int swap,l,ntic;
	float size,dv,angle;
	float t;

			/* offset bounding box to physical space */
	x1 -= pxmin;
	x2 -= pxmin;
	y1 -= pyminb;
	y2 -= pyminb;

			/* determine if need to swap end points to 
			   make sure that annotation appears on 
			   correct side of axis */
	swap = 0;
	if(type < 2){
		l = -999;
		if(x1 > x2) swap = 1;
	}else{
		l = 999;
		if(y1 > y2) swap = 1;
	}

			/* swap end point if ncessary */
	if(swap){
		t = x1;
		x1 = x2;
		x2 = t;
		t = y1;
		y1 = y2;
		y2 = t;
		dir = -dir;	/* dir determines annotation swap */
	}

			/* compute number of tick marks and return if
			axis is too small */
	size = hypot(x1-x2,y1-y2);
	ntic = size + 1.0;
	if(ntic <= 1) return;

			/* swap annotation if necessary */
	if(dir < 0){
		t = vmax;
		vmax = vmin;
		vmin = t;
	}

			/* get angle in degrees */
	angle = atan2(y2-y1,x2-x1) * 180.0 / M_PI;

			/* figure step for axis */
	dv = (vmax-vmin) / ( ntic - 1);

	axis(x1,y1,label,l,-size,angle,vmin,dv);
}

void plot3d(x,y,z,nx,ny,xmin,xmax,ymin,ymax,zmin,zmax,
xlen,ylen,zlen,phi1,phi2,resol,pxaxis,
pyaxis,pzaxis,xdir,ydir,tl,bl,xl,yl,zl)

float *x;		/* x-data vector; length nx */
float *y;		/* y-data vector; length ny */
float *z;		/* z-matrix; size nx*ny */
int nx;			/* size of xd and zd */
int ny;			/* size of yd and zd */

			/* min and max values for xd, yd, and zd.
			they can have larger, but not smaller, extent
			than the acutal data. the x and y vectors have
			to contain data that are monotonically increasing
			from their min values to their max values. */
double xmin,xmax;	
double ymin,ymax;
double zmin,zmax;

			/* ... target length of three axes. the actual
			length depends on the rotation angles.  in one
			step a bounding box is given by the axis
			lengths specified. the box is rotated by ph1
			and phi2 (see below) and then projected onto
			the physical plotting area.  the projection is
			circumscribed with a rectangle whose sides are
			parallel to the physical x and y axes. the
			lower left corner of rectangle is placed at the
			present plotting origin (see plot(3)). the
			function does not change the origin nor the
			scale factor during plottting. it is the user's
			responsibility to make sure that the graph will
			fit in the available plotting area given the
			origin and scale factor. clearly the axes
			lengths are a key to the sizing.  */
double xlen,ylen,zlen;

			/* orientation of x-y plane in degrees. 
			this is a first rotation and is about z. */
double phi1;		

			/* tilt of plane in degrees.
			this is a second rotation and is about x */
double phi2;		

			/* plotting resolution. determines steps for 
			line hiding. value controls quality of output.
			values of about one are good for preview. values
			greater than about two are required for 
			postscript quality */ 
double resol;		

			/* 1=plot specified axis */
int pxaxis,pyaxis,pzaxis;

int xdir,ydir;		/* 1=plot lines parallel to specified axis */
char *tl;		/* top label */
char *bl;		/* bottom label */
char *xl,*yl,*zl;	/* axes labels */
{
	int i,j,index1,index2;	/* indexes for for loops */
	float xfct,yfct,zfct;	/* scale from object to physical space */
	float q;		/* temp floating point value */
	int x1,x2,y1,y2;	/* elevation locations */
	float dmin;		/* temp floating point value */
	int xs=-1,xe=-1,ys=-1,ye=-1;	/* used to determine nearest corner */
	int corner;		/* corners to key plots */
	int axcor,aycor,azcor;	/* corners for each axis */
	int axdir=-1,aydir=-1;	/* direction to plot axes */
	int xtype=-1;		/* where x axis is plotted */
	float dm;		/* scale factor to discrete space */

				/* projections of corners of bounding box */
	float corx[4],cory[8],cord[4];

				/* bounds for for loops that are used to 
				   step through elevation array */
	int istep1,istep2,istrt1,istrt2,iend1,iend2;
/*
	check for open plotter
*/
	plots_open_check("plot3d");
/*
	check parameters and data
*/
	if(nx < 2 || ny < 2)
		err("nx and ny must be greater than one");

	for(i=0;i<nx;i++) if(x[i] > xmax || x[i] < xmin)
		err("x data out of range");

	for(i=0;i<ny;i++) if(y[i] > ymax || y[i] < ymin)
		err("y data out of range");

	if(xlen <= 0 || ylen <= 0 || zlen <= 0)
		err("improper axis length");

	if(xmin == xmax) err("no x range");
	if(ymin == ymax) err("no y range");
	if(zmin == zmax) err("no z range");
/*
	flip x and y vectors when monotonically decreasing
*/
	i = monotonicity(x,nx);
	if(i == 2 || i == 4){
		q = xmax;
		xmax = xmin;
		xmin = q;
	}

	i = monotonicity(y,ny);
	if(i == 2 || i == 4){
		q = ymax;
		ymax = ymin;
		ymin = q;
	}
/*
	setup rotation factors for use in xcal, ycal and dcal
*/
	phi1 *= M_PI / 180.0;
	phi2 *= M_PI / 180.0;
	sphi1 = sin(phi1);
	cphi1 = cos(phi1);
	sphi2 = sin(phi2);
	cphi2 = cos(phi2);
	s1s2 = sphi1 * sphi2;
	c1s2 = cphi1 * sphi2;
	c2s1 = cphi2 * sphi1;
	c1c2 = cphi2 * cphi1;
/*
	set global coordinate transformation parameters to object space
*/
	absxs = 0.0;
	absys = 0.0;
/*
	project bounding box
*/
	corx[0] = xcal(0.0,0.0);
	cord[0] = dcal(0.0,0.0,0.0);
	cory[0] = ycal(0.0,0.0,0.0);
	cory[4] = ycal(0.0,0.0,zlen);

	corx[1] = xcal(xlen,0.0);
	cord[1] = dcal(xlen,0.0,0.0);
	cory[1] = ycal(xlen,0.0,0.0);
	cory[5] = ycal(xlen,0.0,zlen);

	corx[2] = xcal(xlen,ylen);
	cord[2] = dcal(xlen,ylen,0.0);
	cory[2] = ycal(xlen,ylen,0.0);
	cory[6] = ycal(xlen,ylen,zlen);

	corx[3] = xcal(0.0,ylen);
	cord[3] = dcal(0.0,ylen,0.0);
	cory[3] = ycal(0.0,ylen,0.0);
	cory[7] = ycal(0.0,ylen,zlen);
/*
	look at plotting volume and find closest corner
*/
	dmin = cord[0];
	corner = 0;
	for(i=1;i<4;i++){
		if(cord[i] < dmin){
			dmin = cord[i];
			corner = i;
		}
	}
	if(debug) printf("dmin corner: %d\n",corner);
/*
	set limits on for loops to determine how to step through data

	corners set according to following diagram:

	y-axis

	|3=(0,ylen)  2=(xlen,ylen)
	|
	|
	|
	|
	|
	|------------- x axis
	0=(0,0)       1=(xlen,0)
*/
	switch(corner){

	case 0:
		xs = 0; 
		xe = nx;
		ys = 0; 
		ye = ny;
		break;

	case 1:
		xs = nx -1; 
		xe = -1;
		ys = 0; 
		ye = ny;
		break;

	case 2:
		xs = nx -1; 
		xe = -1;
		ys = ny - 1; 
		ye = -1;
		break;

	case 3:
		xs = 0; 
		xe = nx;
		ys = ny - 1; 
		ye = -1;
		break;
	}

	if(ydir){
		istrt1 = xs;
		iend1 = xe;
		istrt2 = ys;
		iend2 = ye;
	}else{
		istrt1 = ys;
		iend1 = ye;
		istrt2 = xs;
		iend2 = xe;
	}

	istep1 = (iend1 > istrt1)? 1 : -1;
	istep2 = (iend2 > istrt2)? 1 : -1;
/*
	look at plotting volume and find bottom corner for x and y axes
*/
	dmin = cory[0];
	corner = 0;
	for(i=1;i<4;i++){
		if(cory[i] < dmin){
			dmin = cory[i];
			corner = i;
		}
	}
	if(debug) printf("ymin corner: %d\n",corner);
/*
	set relative arrangement of axes
*/
	axcor = aycor = -1;	/* to bypass compiler warning */

	switch(corner){

	case 0:
		axdir = 1;
		aydir = -1;
		xtype = 1;

		axcor = 0;
		aycor = 3;
		break;

	case 1:
		axdir = 1;
		aydir = 1;
		xtype = 0;
		axcor = 0;
		aycor = 1;
		break;

	case 2:
		axdir = -1;
		aydir = 1;
		xtype = 1;
		axcor = 2;
		aycor = 1;
		break;

	case 3:
		axdir = -1;
		aydir = -1;
		xtype = 0;
		axcor = 2;
		aycor = 3;
		break;
	}

/*
	look at plotting volume to see where to put x axes
*/
	dmin = corx[0];
	azcor = 0;
	for(i=1;i<4;i++){
		if(corx[i] < dmin){
			dmin = corx[i];
			azcor = i;
		}
	}
	if(debug) printf("axcor, aycor, azcor: %d %d %d\n",axcor,aycor,azcor);
/*
	find total extent of projection of bounding box
*/
	for(pxmax=pxmin=corx[0],i=1;i<4;i++){
		if(corx[i] > pxmax) pxmax = corx[i];
		if(corx[i] < pxmin) pxmin = corx[i];
	}

	for(pymaxb=pyminb=cory[0],i=1;i<8;i++){
		if(cory[i] < pyminb) pyminb = cory[i];
		if(cory[i] > pymaxb) pymaxb = cory[i];
	}
/*
	set scale factors from object to plotting space
*/
	xfct = xlen / (xmax - xmin);
	yfct = ylen / (ymax - ymin);
	zfct = zlen / (zmax - zmin);
/*		
	number of steps for line hiding		
*/
	numstp = 2048 * resol;
	if(numstp < 128)err("not enough internal resolution");
	dm = (numstp-1) / (pxmax - pxmin);
	iplot_scale = 1.0 / dm;
/*
	get malloc'ed space for internal vectors
*/
	i = (nx > ny)? nx : ny;
	if(NULL == (absx = (int *) malloc((i*4+numstp*2)*sizeof(int)))){
		err("malloc failure\n");
	}
	absy = absx + i;
	xold = absx + 2 * i;
	yold = absx + 3 * i;
	above = absx + 4 * i;
	below = absx + 4 * i + numstp;
/*
	init upper horizon array 
*/
	for(i=0;i<numstp;i++) above[i] = ZINIT;
/*
	set global coordinate transformation parameters to plotting space
*/
	cphi1 *= xfct;
	sphi1 *= yfct;
	absxs = xmin * cphi1 + ymin * sphi1 + pxmin;

	s1s2 *= xfct;
	c1s2 *= yfct;
	cphi2 *= zfct;
	absys = -s1s2 * xmin + c1s2 * ymin + cphi2 * zmin + pyminb;
/*
	get physical ymax from actual data
*/
	pymin = pymax = ycal(x[0],y[0],Z(0,0));
	for(i=0;i<nx;i++){
		for(j=0;j<ny;j++){
			q = Z(i,j);
			if(q < zmin || q > zmax) err("z data out of range");
			q = ycal(x[i],y[j],q);
			pymax = max(q,pymax);
			pymin = min(q,pymin);
		}
	}
/*
	draw top and bottom labels
*/
	if(tl && (i=strlen(tl)) > 0){
		symbol((pxmax-pxmin)/2-HEIGHT*i/2,
			pymax+3*HEIGHT,HEIGHT,tl,0.0,i);
	}

	if(bl && (i=strlen(bl)) > 0){
		symbol((pxmax-pxmin)/2-HEIGHT*i/2,-5*HEIGHT,
		  HEIGHT,bl,0.0,i);
	}
/*		
	draw axes
*/
	if(pzaxis){
		j = azcor;
		drawaxes(
			corx[j],
			cory[j],
			corx[j],
			cory[j+4],
			zmin,
			zmax,
			1,
			2,
			zl);
	}

	if(pxaxis){
		j = (axcor + 1) % 4;
		drawaxes(
			corx[axcor],
			cory[axcor],
			corx[j],
			cory[j],
			xmin,
			xmax,
			axdir,
			xtype,
			xl);
	}

	if(pyaxis){
		j = (aycor + 1) % 4;
		drawaxes(
			corx[aycor],
			cory[aycor],
			corx[j],
			cory[j],
			ymin,
			ymax,
			aydir,
			!xtype,
			yl);
	}
/*
	main plotting loop				
*/
	for(index1=istrt1;index1!=iend1;index1+=istep1){
/*   		
	calculate the line coordinates for next row
*/
	for(i=istrt2;i!=iend2;i+=istep2){
		if(ydir){
			absx[i] = dm * xcal(x[index1],y[i]);
			absy[i] = dm * ycal(x[index1],y[i],Z(index1,i));
		}else{
			absx[i] = dm * xcal(x[i],y[index1]);
			absy[i] = dm * ycal(x[i],y[index1],Z(i,index1));
		}
	}
					
	if(index1 != istrt1 ){ 		/* no need for last row */
		if(xdir && ydir){
			for(index2=istrt2;index2!=iend2;index2+=istep2){
				perpen(index2);
			}
		}else{
			perpen(istrt2);
		}
	}
/*		
	loop for drawing one complete line  		
*/
	x2 = absx[istrt2];
	y2 = absy[istrt2];

	for(index2=istrt2;index2!=iend2-istep2;index2+=istep2){
		x1 = x2;
		y1 = y2;
		x2 = absx[index2+istep2];
		y2 = absy[index2+istep2];
		drawline(x1,y1,x2,y2);
	}
	if((index1 != istrt1) && !(xdir && ydir)) perpen(iend2-istep2);
/*			
	store the line just drawn			
*/
	for(i=istrt2;i!=iend2;i+=istep2){
		xold[i] = absx[i];
		yold[i] = absy[i];
	}
/*		
	end of main loop				
*/
	}
/*
	clean up and return
*/
	free(absx);
	return;
}


/*
program notes:

drawline is the key function that performs hidden line removal.  it
works by keeping track of an upper and a lower horizon at points along
the physical x-axis. each new line segment is compared against the
existing horizons and the visible portions of the new segment are
displayed and the horizons are updated.  ideally a linked list of
segments for each horizon should be maintained. instead, the
horizontal axis is segmented into discrete steps and the horizons are
maintained in arrays corresponding to the discrete steps. the number of
steps determines the quality of the resulting graphics. the number of
steps should be on the order of the resolution of the display device.

the coordinates of the line segments are specified in terms of the
discrete space on which the horizons are sampled. the function iplot as
an integer interface to plot(3) and handles the scaling back to
physical space. there is some wasted calculations here because plot(3)
already does a scale multiplication and origin offset.  the
calculations are duplicated to avoid having to reset plotting
parameters.

the bresenham algorithm is used to walk along the elevation (y) at each
x sample of each segment. at any given time, you know if you are above,
below or on each of the two horizons. for lines with (absolute) slopes
less than one a state transition table can be generated to figure out
what lines to draw given the state of a current sample and the state of
the previous sample. for example, if a line starts out above the upper
horizon, we walk along it until it crosses the upper horizon. when it
crosses, we draw a line from the start to the intersection. also, the
horizon is updated to reflect that the new segment is above the old
horizon.  if there is no intersection, then all of the segment is above
the horizon and the line is drawn when the function exists.

the use of sampled horizon vectors fails for segments with (absolute)
slopes greater than one. these are called steep lines. the problem is
that at a given x, more than one y can be examined. a good example is
to look at a cliff with a vertical face. the segment corresponding to
the face is below the upper horizon. the use of the algorithm described
in the previous section would not allow the face to be drawn.
therefore, extra cases are inserted into the state table to handle the
steep lines. because of time contraints, the cases were added
heuristically until the graphs looked correct. at high magnifications,
you can still see some problems. perhaps the only way to avoid the
problems is to convert to a linked list.  a number of problems might be
related to the use of ">" and "<" instead of "<=" and ">=" and visa
versa. the use of forshortening (see a standard text on graphics) might
help.  a main problem is how to set the horizon on a vertical line.  at
high magnifications you can see that lines do not intersect these
verticals at consistent points.

the original version of the code would initialize the upper and lower
horizons to large negative and positive numbers, respectively.  it was
therefore possible for segments to be simultaneously above the upper
horizon and below the lower horizon. while physically impossible, the
state table was expanded to handle this case so that the horizons
didn't have to be filled in with real values. in the present version
the upper horizon is filled in with a large value, defined by ZINIT,
and called a key. the code checks for this value. the state table is
simplified with the use of the key. there might be some code remaining
from the old state table. however, due to the tricks required to
implement steep lines, it was thought to be better to leave the extra
code rather than break the tricks.
*/
